/*
  PSthreads and UthLib thread libraries
  == Copyright (C) 2007 the OMPi/PSthreads Team
  == Department of Computer Science, University of Ioannina

  This file is part of PSthreads.

  PSthreads is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  PSthreads is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with PSthreads; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*! \file psthread.c
    \brief PSthreads runtime library
 */

#include "common.h"
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <uth.h>

typedef union _vp_psthread {
	struct {
		psthread_t *__ps_unlock_thread;		/* Thread to have its struct unlocked */
		psthread_t *__ps_yield_thread;		/* Thread to have its struct unlocked */
		psthread_t *__ps_enqueue_thread;	/* Thread to be enqueued */
		psthread_t *__ps_scheduler_thread;	/* Scheduler thread for this vp */
	} thr;
	char pad[CACHE_LINE_SIZE];
} vp_psthread_t;

struct _ps_global_data {
	vp_psthread_t	__ps_vp_psthr[MAX_VPS];

	queue_t			__ps_local_ready_queue[MAX_VPS];
	queue_t			__ps_lreuse_queue[MAX_VPS];
	queue_t			__ps_ready_queue;
	queue_t 		__ps_reuse_queue;

	pthread_t		__ps_vpid[MAX_VPS];
} *_ps_gdata;

#define _ps_unlock_thread(x)	_ps_gdata->__ps_vp_psthr[x].thr.__ps_unlock_thread
#define _ps_yield_thread(x)		_ps_gdata->__ps_vp_psthr[x].thr.__ps_yield_thread
#define _ps_enqueue_thread(x)	_ps_gdata->__ps_vp_psthr[x].thr.__ps_enqueue_thread
#define _ps_scheduler_thread(x)	_ps_gdata->__ps_vp_psthr[x].thr.__ps_scheduler_thread

#define _ps_local_ready_queue	_ps_gdata->__ps_local_ready_queue
#define _ps_ready_queue			_ps_gdata->__ps_ready_queue
#define _ps_lreuse_queue		_ps_gdata->__ps_lreuse_queue
#define _ps_reuse_queue			_ps_gdata->__ps_reuse_queue

#define _ps_vpid			_ps_gdata->__ps_vpid


int			_ps_nvps;
volatile int		vp_created = 1;


static psthread_attr_t default_attr = PSTHREAD_ATTR_INITIALIZER;

/* forward declarations */
int psthread_create (psthread_t **t, psthread_attr_t *attr, void (*start)(void *arg), void *arg);
/*inline*/ psthread_t *psthread_self();
/*inline*/ psthread_t *psthread_self_ex(int vp);
static void _ps_thread_main(void *arg);


/*
 * High-level lock primitives 
 */

void psthread_lock_init(psthread_lock_t *spin)
{
	_lock_init(spin);
}

void psthread_lock_acquire(psthread_lock_t *spin)
{
	_lock_acquire(spin);
}

int psthread_lock_try_acquire(psthread_lock_t *spin)
{
	return _lock_try_acquire(spin);
}

void psthread_lock_release(psthread_lock_t *spin)
{
	_lock_release(spin);
}

void psthread_lock_destroy(psthread_lock_t *spin)
{
	_lock_destroy(spin);
}

/*
 * Implementation of the body of the scheduling loop. It is continuously
 * executed by each virtual processor. It does some cleanup work, selects
 * the next user-level thread to be executed and context-switches to it.
 */ 

static inline psthread_t *stealing(int vp)
{  
	int i;
	psthread_t *thread = NULL;            

	i = (vp + 1) % _ps_nvps;
	while ((thread == NULL) && (i != vp)) {
		_dequeue_end(&_ps_local_ready_queue[i], &thread);
		i = (i + 1) % _ps_nvps;
		break;
	}                  

	return thread;
}

static inline void _ps_vp_schedule(long vp)
{
	psthread_t *thread = NULL;
	psthread_t *unlock_thr = NULL;

	unlock_thr = _ps_unlock_thread(vp);
	if (unlock_thr != NULL) {
		_ps_unlock_thread(vp) = NULL;

		_lock_release(&unlock_thr->lock);	
		if (unlock_thr->state == _PS_ZOMBIE) {
			uth_delete(unlock_thr->context);
#if defined(UTH_REUSE)
#if defined(UTH_LOCAL_REUSE)
			_enqueue_head_bare(&_ps_lreuse_queue[vp], unlock_thr);
#else
			_enqueue_head(&_ps_reuse_queue, unlock_thr);
#endif
#endif
		}
	}

	if (_ps_yield_thread(vp) != NULL) {
#if defined(PS_GLOBAL_QUEUE_ONLY)
		_enqueue_tail(&_ps_ready_queue, _ps_yield_thread(vp));
#else
		_enqueue_tail(&_ps_local_ready_queue[vp], _ps_yield_thread(vp));
#endif
		_ps_yield_thread(vp) = NULL;
	}

	if (_ps_enqueue_thread(vp) != NULL) {
		thread = _ps_enqueue_thread(vp);	/* nice option but not appropriate for yield */
		_ps_enqueue_thread(vp) = NULL;
	}

	if (thread == NULL) _dequeue(&_ps_ready_queue, &thread);	/* coarse grain */

#if !defined(PS_GLOBAL_QUEUE_ONLY)
	if (thread == NULL) _dequeue(&_ps_local_ready_queue[vp], &thread);	/* check locally */

#if defined(PS_STEALING)
	if (thread == NULL) thread = stealing(vp);
#endif

#endif

	if (thread == NULL) {
		thread = _ps_scheduler_thread(vp);
		thread->state = _PS_RUNNING;
#if defined(PS_YIELDVP)
		sched_yield();
#endif
		return;
	}

	/* Resume the thread */
	thread->state = _PS_RUNNING;
	thread->vpid = vp;
	if (thread->context == NULL) {
		thread->context = uth_create(_ps_thread_main, thread);
	}

	uth_switchto_ex(vp, _ps_scheduler_thread(vp)->context, thread->context);
}

/* 
 * Voluntarily release your virtual processor
 */
void psthread_yield()
{
	int vp = uth_get_vpid();
	psthread_t *me = psthread_self_ex(vp);

	me->state = _PS_RUNNABLE;
	_ps_yield_thread(vp) = me;
	uth_switchto_ex(vp, me->context, _ps_scheduler_thread(vp)->context);
}


/*
 * Scheduling loop
 */

void _ps_scheduler_thread_start(void *arg)
{
	long vp = (long)arg;

	while (1) {
		_ps_vp_schedule(vp);
	}
}

/*
 * Initialization of virtual processor
 */
static void *_ps_init_vp(void *arg)
{
	long vp = (long) arg;
	psthread_t *thread;
	
	uth_vp_init(vp,0);

	/*
	 * Initialize primordial thread for this virtual processor
	 */
	
	thread = (psthread_t *) _ort_calloc_aligned(sizeof(psthread_t), NULL);
	if (!thread)
		return (void *) EAGAIN;
	
	thread->state = _PS_RUNNING;
	thread->nchildren = -1;
	thread->context = uth_self_ex(vp);	/* uth_vp_init has already assigned an underlying thread (uth) to the current virtual processor  */
	uth_setarg(uth_self_ex(vp), thread);	/* associate this user level thread with the corrersponding uth */
	_lock_init(&thread->lock);
	thread->vpid = vp;
	
	/*
	 * Create scheduler thread
	 */
	
	psthread_create(&_ps_scheduler_thread(vp), &default_attr, _ps_scheduler_thread_start, (void *) vp);

	if (!_ps_scheduler_thread(vp))
		return NULL;
	
	_ps_scheduler_thread(vp)->context = uth_create(_ps_thread_main, _ps_scheduler_thread(vp));
	
	vp_created++;

	uth_switchto_ex(vp, NULL, _ps_scheduler_thread(vp)->context);
	
	/* Never returns here */
	return 0;
}


/*
 * Initializes the library structures, creates virtual processors and
 * scheduler threads.
 */
int psthread_init(int vps, unsigned long stacksize)
{
	psthread_t *thread;
	long i;
	pthread_t dummy;
	pthread_attr_t attr;
	
/*	printf("psthread_init(%d, %ld) called\n", vps, stacksize);*/
	
	_ps_gdata = (struct _ps_global_data *) _ort_calloc_aligned(sizeof(struct _ps_global_data), NULL);
	_ps_nvps = vps;
	_queue_init(&_ps_ready_queue);
	_queue_init(&_ps_reuse_queue);
	for (i = 0; i < vps; i++) {
		_queue_init(&_ps_local_ready_queue[i]);
		_queue_init(&_ps_lreuse_queue[i]);
	}

	
	uth_init(stacksize);	/* Initialize uthlib */
	uth_vp_init(0,0);	/* Initialize the first virtual processor (Posix Thread) */

	/*
	 * Initialize primordial thread for this virtual processor
	 */
	
	thread = (psthread_t *) _ort_calloc_aligned(sizeof(psthread_t), NULL);
	if (!thread)
		return EAGAIN;
	
	thread->state = _PS_RUNNING;
	thread->nchildren = -1;
	thread->context = uth_self_ex(0);	/* uth_vp_init has already assigned an underlying thread (uth) to the current virtual processor  */
	uth_setarg(uth_self_ex(0), thread);	/* associate this user level thread with the corrersponding uth */
	_lock_init(&thread->lock);
	thread->vpid = 0;


	/*
	 * Create scheduler thread
	 */
	psthread_create(&_ps_scheduler_thread(0), &default_attr, _ps_scheduler_thread_start, (void *) 0);

	if (!_ps_scheduler_thread(0))
		return EAGAIN;
	
	_ps_scheduler_thread(0)->context = uth_create(_ps_thread_main, _ps_scheduler_thread(0));
	
	/* 
	 * Create the rest virtual processors 
	 */
	if (vps > 1) {
		pthread_attr_init(&attr);
		if (pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM) != 0) {
			fprintf(stderr, "[PSTHREADS] warning: pthread_attr_setscope failed\n");
			fflush(stderr);
		}

		if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) != 0) {
			fprintf(stderr, "[PSTHREADS] warning: pthread_attr_setdetachstate failed\n");
			fflush(stderr);
		}
		
		for (i = 1; i < vps; i++) {
			if (pthread_create(&dummy, &attr, _ps_init_vp, (void *)i) != 0) {
				fprintf(stderr, "[PSTHREADS] error: pthread_create failed\n");
				fflush(stderr);
				exit(1);
			}
		}

		while (vp_created < vps) {
			sched_yield();
		}
		
	}

	return 0;
}

/* 
 * Exit function called by each user-level thread for normal termination.
 * It never returns.
 */
static inline void _psthread_exit(psthread_t *me)
{
	int vp = me->vpid;
	int res;

	me->state = _PS_ZOMBIE;

	/* If I am joinable */
	if (me->detachstate == PSTHREAD_JOINABLE) {	
		_lock_acquire(&me->parent->lock);
		res =  (--me->parent->nchildren == 0) && (me->parent->state == _PS_WAITING);
		_lock_release(&me->parent->lock);
		/* If my parent waits and I am its last child, make it runnable */
		if (res) {
			me->parent->state = _PS_RUNNABLE;
			_ps_enqueue_thread(vp) = me->parent;
		}
	}

	_ps_unlock_thread(vp) = me;

	uth_switchto_ex(vp, NULL, _ps_scheduler_thread(vp)->context);
}


void psthread_exit()
{
	int vp= uth_get_vpid();
	psthread_t *me = psthread_self_ex(vp);

	_psthread_exit(me);
}

/* 
 * Driver function. It is called by each newly created user-level thread.
 * It executes the user-provided work function and ensures proper termination
 */
static void _ps_thread_main(void *arg)
{
	psthread_t *thread = arg;
	
	/* Run thread main */
	(*thread->start)(thread->arg);
	
	/* All done, time to go away */
	_psthread_exit(thread);
}



/* 
 * Thread creation function: it allocates the thread structure
 * and creates the corresponding underlying thread.
 */
int psthread_create(psthread_t **t, psthread_attr_t *attr, void (*start)(void *arg), void *arg)	
{
	psthread_t *me;
	int res = 0;
	psthread_t *thread = NULL;

	if ((attr == NULL) || (attr->parent == NULL)) {
		me = psthread_self();
	}
	else {
		me = attr->parent;
	}

#if defined(UTH_REUSE)
#if defined(UTH_LOCAL_REUSE)
	_dequeue_bare(&_ps_lreuse_queue[me->vpid], &thread);
#else
	_dequeue(&_ps_reuse_queue, &thread);
#endif
#endif
	if (thread == NULL) {
		thread = _ort_calloc_aligned(sizeof(psthread_t), NULL);
	}

	if (thread == NULL) {
		return EAGAIN;
	}

	/* Initialize thread */
	thread->start = start;
	thread->arg = arg;
	_lock_init(&thread->lock);
	thread->prev = thread->next = NULL;
	thread->vpid = -1;	
	thread->nchildren = 0;
	thread->detachstate= (attr == NULL)? 0: attr->detachstate;
	thread->parent = me;

	/* Any psthread is mapped to an underlying thread.
	 * It is executed through a driver routine that takes
	 * as argument the psthread itself
	 */
	thread->context = NULL;
	
	/* Make thread runnable */
	thread->state = _PS_NEW;
	
	if (thread->detachstate == PSTHREAD_JOINABLE) {
		/* Increase my children */
		_lock_acquire(&me->lock);
		me->nchildren++;
		_lock_release(&me->lock);
	}

	/* Store the new thread descriptor */
	*t = thread;


	return res;
}


/* 
 * Parent thread waits for all its children to terminate 
 */
int psthread_waitall()
{
	int vp = uth_get_vpid();
	psthread_t *me = psthread_self_ex(vp);

	_lock_acquire(&me->lock);
	if (me->nchildren == 0) {
		_lock_release(&me->lock);
		return 1;
	}
	me->state = _PS_WAITING;
	_ps_unlock_thread(vp) = me;
	uth_switchto_ex(vp, me->context, _ps_scheduler_thread(vp)->context);

	return 0;
}

/*
 * Queue management routines
 */
int psthread_enqueue_tail(psthread_t *thread, int queue)
{
	if ((queue < -1) && (queue >= _ps_nvps))
		return EINVAL;

	thread->state = _PS_RUNNABLE;

#if defined(PS_GLOBAL_QUEUE_ONLY)
	queue = -1;
#endif
	if (queue == -1) {
		_enqueue_tail(&_ps_ready_queue, thread);
	} 
	else {
		_enqueue_tail(&_ps_local_ready_queue[queue], thread);
	}

	return 0;
}

int psthread_enqueue_head(psthread_t *thread, int queue)
{
	if ((queue < -1) && (queue >= _ps_nvps))
		return EINVAL;

	thread->state = _PS_RUNNABLE;

#if defined(PS_GLOBAL_QUEUE_ONLY)
	queue = -1;
#endif

	if (queue == -1) {
		_enqueue_head(&_ps_ready_queue, thread);
	} 
	else {
		_enqueue_head(&_ps_local_ready_queue[queue], thread);
	}

	return 0;
}

int psthread_enqueue(psthread_t *thread, int queue)
{
	return psthread_enqueue_tail(thread, queue);
}


/* 
 * Return a pointer to the descriptor of the calling thread
 */ 
psthread_t *psthread_self(void)
{
	/* The descriptor is the argument of the corresponding underlying thread */
	int vp = uth_get_vpid();
	return uth_getarg_ex(vp);
}

psthread_t *psthread_self_ex(int vp)
{
	/* The descriptor is the argument of the corresponding underlying thread */
	return uth_getarg_ex(vp);
}


/* 
 * Return elapsed time in sec
 */
double psthread_gettime()
{
	return uth_gettime();
}

/* 
 * Suspend virtual processor for ms milliseconds
 */
void psthread_vp_sleep(int ms)
{
	uth_vp_sleep(ms);
}


/*
 * Virtual processors info
 */
int psthread_nvps(void)
{
	return _ps_nvps;
}

int psthread_current_vp(void)
{
	return uth_get_vpid();
}

unsigned long psthread_vp_self()
{
	return uth_vp_self();
}


/*
 * Condition variables
 */

void psthread_cond_init(psthread_cond_t *cond)
{
	_queue_init(cond);
}

void psthread_cond_destroy (psthread_cond_t *cond)
{
	_queue_destroy(cond);
}

void psthread_cond_wait (psthread_cond_t *cond, psthread_lock_t *lock)
{
	int vp = uth_get_vpid();
	psthread_t *me = psthread_self_ex(vp);

	if (lock != NULL) {
		_lock_acquire(&me->lock);
	}

	_enqueue_tail(cond, me);

	if (lock != NULL) {
		_lock_release(lock);
	}

	me->state = _PS_WAITING;
	_ps_unlock_thread(vp) = me;

	uth_switchto_ex(vp, me->context, _ps_scheduler_thread(vp)->context);
	if (lock != NULL) {				/* for barrier */
		_lock_acquire(lock);
	}
}

void psthread_cond_signal (psthread_cond_t *cond)
{
	psthread_t *thread = NULL;

	_dequeue(cond, &thread);
	if (thread != NULL) {
		_lock_acquire(&thread->lock);
		thread->prev = NULL;
		thread->next = NULL;
		thread->state = _PS_RUNNABLE;
		_lock_release(&thread->lock);
#if defined(PS_GLOBAL_QUEUE_ONLY)
		_enqueue_head(&_ps_ready_queue, thread);
#else
		_enqueue_head(&_ps_local_ready_queue[thread->vpid], thread);
#endif
	}
	else {
		_queue_init(cond);	/* is it necessary ? */
	}
}

void psthread_cond_broadcast (psthread_cond_t *cond)
{
	psthread_t *thread, *next;

	thread = _queue_head(cond);
	_queue_init(cond);
	while (thread != NULL) {
		_lock_acquire(&thread->lock);
		next = thread->next;
		thread->prev = thread->next = NULL;
		thread->state = _PS_RUNNABLE;
		_lock_release(&thread->lock);
#if defined(PS_GLOBAL_QUEUE_ONLY)
		_enqueue_head(&_ps_ready_queue, thread);
#else
		_enqueue_head(&_ps_local_ready_queue[thread->vpid], thread);
#endif
		thread = next;
	}
}


/*
 * Thread Local Storage (not fully functional)
 */
 
static int tls_keys[MAX_TLS_KEYS] = {0};
static pthread_mutex_t tls_mutex = PTHREAD_MUTEX_INITIALIZER;

int psthread_key_create(psthread_key_t *key, void *destructor)
{
	int i;

	pthread_mutex_lock(&tls_mutex);
	for (i = 0; i < MAX_TLS_KEYS; i++) {
		if (tls_keys[i] == 0) {
			tls_keys[i] = 1;
			*key = i;
			pthread_mutex_unlock(&tls_mutex);
			return 0;
		}
	}
	pthread_mutex_unlock(&tls_mutex);
	*key = -1;
	return 1;
}

int psthread_setspecific(psthread_key_t key, const void *value)
{
	int vp = uth_get_vpid();
	psthread_t *t = uth_getarg_ex(vp);

	if ((key < 0) || (key >= MAX_TLS_KEYS)) {
		return 1;
	}
	t->tls[key] = value;
	return 0;
}

void *psthread_getspecific(psthread_key_t key)
{
	int vp = uth_get_vpid();
	psthread_t *t = uth_getarg_ex(vp);

	if ((key < 0) || (key >= MAX_TLS_KEYS)) {
		return NULL;
	}
	return (void *)t->tls[key];
}

int psthread_key_delete(psthread_key_t key)
{
	if ((key < 0) || (key >= MAX_TLS_KEYS)) {
		return 1;
	}
	pthread_mutex_lock(&tls_mutex);
	tls_keys[key] = 0;
	pthread_mutex_unlock(&tls_mutex);

	return 0;
}


/*
 * Posix-like barriers (D. Butenhof)
 */

/*
 * Initialize a barrier for use.
 */
int psthread_barrier_init (psthread_barrier_t *barrier, int count)
{
	barrier->threshold = barrier->counter = count;
	barrier->cycle = 0;
	if (barrier->valid != _PS_BARRIER_VALID) {
		_lock_init (&barrier->mutex);
		psthread_cond_init (&barrier->cv);	/* _queue_init(&barrier->cv); */
		barrier->valid = _PS_BARRIER_VALID;
	}
	return 0;
}

/*
 * Destroy a barrier when done using it.
 */
int psthread_barrier_destroy (psthread_barrier_t *barrier)
{
	if (barrier->valid != _PS_BARRIER_VALID)
		return EINVAL;

	_lock_acquire(&barrier->mutex);

	/*
	 * Check whether any threads are known to be waiting; report	
	 * "BUSY" if so.
	 */
	if (barrier->counter != barrier->threshold) {
		_lock_release (&barrier->mutex);
		return EBUSY;
	}

	barrier->valid = 0;
	_lock_release(&barrier->mutex);

	/*
	 * If unable to destroy either 1003.1c synchronization
	 * object, return the error status.
	 */

	_lock_destroy (&barrier->mutex);
	psthread_cond_destroy (&barrier->cv);	/* _queue_destroy(&barrier->cv); */

	return 0;
}

/*
 * Wait for all members of a barrier to reach the barrier. When
 * the count (of remaining members) reaches 0, broadcast to wake
 * all threads waiting.
 */
int psthread_barrier_wait (psthread_barrier_t *barrier)
{
	int status;
	unsigned long cycle;

	if (barrier->valid != _PS_BARRIER_VALID)
		return EINVAL;

	_lock_acquire(&barrier->mutex);

	cycle = barrier->cycle;   /* Remember which cycle we're on */

	if (--barrier->counter == 0) {

		barrier->cycle++;
		barrier->counter = barrier->threshold;

		psthread_cond_broadcast (&barrier->cv);

		/*
		 * The last thread into the barrier will return status
		 * -1 rather than 0, so that it can be used to perform
		 * some special serial code following the barrier.
		 */
		status = -1;
	} 
	else {
		/*
		 * Wait until the barrier's cycle changes, which means
		 * that it has been broadcast, and we don't want to wait
		 * anymore.
		 */
		while (cycle == barrier->cycle) {
			psthread_cond_wait (&barrier->cv, &barrier->mutex);
		}
		status = 0;
	}

	_lock_release(&barrier->mutex);

	return status;          /* error, -1 for waker, or 0 */
}
